# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Set of functions to manipulate caches.

Ignite `cache` can be viewed as a named entity designed to store key-value
pairs. Each cache is split transparently between different Ignite partitions.

The choice of `cache` term is due to historical reasons. (Ignite initially had
only non-persistent storage tier.)
"""

from pyignite.connection import Connection
from pyignite.datatypes.cache_config import cache_config_struct
from pyignite.datatypes.primitive import Int, Byte
from pyignite.datatypes.standard import String, StringArray
from pyignite.queries.op_codes import *
from pyignite.queries import Query, Response
from .result import APIResult


def cache_get_configuration(
    conn: Connection, hash_code: int, flags: int=0, query_id=None,
) -> APIResult:
    """
    Gets configuration for the given cache.

    :param conn: connection to Ignite server,
    :param hash_code: hash code of the cache. Can be obtained by applying
     the `hashcode()` function to the cache name,
    :param flags: Ignite documentation is unclear on this subject,
    :param query_id: (optional) a value generated by client and returned as-is
     in response.query_id. When the parameter is omitted, a random value
     is generated,
    :return: API result data object. Result value is OrderedDict with
     the cache configuration parameters.
    """

    class CacheGetConfigurationQuery(Query):
        op_code = OP_CACHE_GET_CONFIGURATION

    query_struct = CacheGetConfigurationQuery([
        ('hash_code', Int),
        ('flags', Byte),
    ], query_id=query_id)

    _, send_buffer = query_struct.from_python({
        'hash_code': hash_code,
        'flags': flags,
    })
    conn.send(send_buffer)

    response_struct = Response([
        ('cache_config', cache_config_struct),
    ])
    response_class, recv_buffer = response_struct.parse(conn)
    response = response_class.from_buffer_copy(recv_buffer)
    result = APIResult(
        status=response.status_code,
        query_id=response.query_id,
    )
    if hasattr(response, 'error_message'):
        result.message = response.error_message
    result.value = response_struct.to_python(response)['cache_config']
    return result


def cache_create(conn: Connection, name: str, query_id=None,) -> APIResult:
    """
    Creates a cache with a given name. Returns error if a cache with specified
    name already exists.

    :param conn: connection to Ignite server,
    :param name: cache name,
    :param query_id: (optional) a value generated by client and returned as-is
     in response.query_id. When the parameter is omitted, a random value
     is generated,
    :return: API result data object. Contains zero status if a cache is
     created successfully, non-zero status and an error description otherwise.
    """

    class CacheCreateQuery(Query):
        op_code = OP_CACHE_CREATE_WITH_NAME

    query_struct = CacheCreateQuery([
        ('cache_name', String),
    ], query_id=query_id)
    _, send_buffer = query_struct.from_python({'cache_name': name})
    conn.send(send_buffer)

    response_struct = Response([])
    response_class, recv_buffer = response_struct.parse(conn)
    response = response_class.from_buffer_copy(recv_buffer)
    result = APIResult(
        status=response.status_code,
        query_id=response.query_id,
    )
    if hasattr(response, 'error_message'):
        result.message = response.error_message
    return result


def cache_get_or_create(conn: Connection, name: str, query_id=None,) -> APIResult:
    """
    Creates a cache with a given name. Does nothing if the cache exists.

    :param conn: connection to Ignite server,
    :param name: cache name,
    :param query_id: (optional) a value generated by client and returned as-is
     in response.query_id. When the parameter is omitted, a random value
     is generated,
    :return: API result data object. Contains zero status if a cache is
     created successfully, non-zero status and an error description otherwise.
    """

    class CacheGetOrCreateQuery(Query):
        op_code = OP_CACHE_GET_OR_CREATE_WITH_NAME

    query_struct = CacheGetOrCreateQuery([
        ('cache_name', String),
    ], query_id=query_id)
    _, send_buffer = query_struct.from_python({'cache_name': name})
    conn.send(send_buffer)

    response_struct = Response([])
    response_class, recv_buffer = response_struct.parse(conn)
    response = response_class.from_buffer_copy(recv_buffer)
    result = APIResult(
        status=response.status_code,
        query_id=response.query_id,
    )
    if hasattr(response, 'error_message'):
        result.message = response.error_message
    return result


def cache_destroy(
    conn: Connection, hash_code: int, query_id=None,
) -> APIResult:
    """
    Destroys cache with a given name.

    :param conn: connection to Ignite server,
    :param hash_code: hash code of the cache. Can be obtained by applying
     the `hashcode()` function to the cache name,
    :param query_id: (optional) a value generated by client and returned as-is
     in response.query_id. When the parameter is omitted, a random value
     is generated,
    :return: API result data object.
    """

    class CacheDestroyQuery(Query):
        op_code = OP_CACHE_DESTROY

    query_struct = CacheDestroyQuery([
        ('hash_code', Int),
    ], query_id=query_id)

    _, send_buffer = query_struct.from_python({
        'hash_code': hash_code,
    })
    conn.send(send_buffer)

    response_struct = Response([])
    response_class, recv_buffer = response_struct.parse(conn)
    response = response_class.from_buffer_copy(recv_buffer)
    result = APIResult(
        status=response.status_code,
        query_id=response.query_id,
    )
    if hasattr(response, 'error_message'):
        result.message = response.error_message
    return result


def cache_get_names(conn: Connection, query_id=None,) -> APIResult:
    """
    Gets existing cache names.

    :param conn: connection to Ignite server,
    :param query_id: (optional) a value generated by client and returned as-is
     in response.query_id. When the parameter is omitted, a random value
     is generated,
    :return: API result data object. Contains zero status and a list of cache
     names, non-zero status and an error description otherwise.
    """

    class CacheGetNamesQuery(Query):
        op_code = OP_CACHE_GET_NAMES

    query_struct = CacheGetNamesQuery(query_id=query_id)

    _, send_buffer = query_struct.from_python()
    conn.send(send_buffer)

    response_struct = Response([
        ('cache_names', StringArray),
    ])
    response_class, recv_buffer = response_struct.parse(conn)
    response = response_class.from_buffer_copy(recv_buffer)
    result = APIResult(
        status=response.status_code,
        query_id=response.query_id,
    )
    if hasattr(response, 'error_message'):
        result.message = response.error_message
    result.value = response_struct.to_python(response)['cache_names']
    return result
